/****
***** Copyright 2010 Intel Corporation All Rights Reserved.
*****
***** The source code, information and material contained herein are owned by Intel Corporation or its suppliers  *****
***** or licensors, and title to such Material remains with Intel Corporation or its suppliers or licensors.      *****
***** The Material contains proprietary information of Intel or its suppliers and licensors. The Material is      *****
***** protected by worldwide copyright laws and treaty provisions. No part of the Material may be used, copied,   *****
***** reproduced, modified, published, uploaded, posted, transmitted, distributed or disclosed in any way without *****
***** Intel's prior express written permission.
*****
***** No license under any patent, copyright or other intellectual property rights in the material is granted to  *****
***** or conferred upon you, either expressly, by implication, inducement, estoppel or otherwise. Any license     *****
***** under such intellectual property rights must be express and approved by Intel in writing.
****/

/**** Copyright Ends ****/

#include <arbb.hpp>


using namespace arbb;


template<typename R, typename T>
void sum_of_squared_differences(R& result, const dense<T>& a, const dense<T>& b)
{
  // copy-conversion c'tor to proceed with higher precision
  const dense<R> c = (a - b) * (a - b);

  // collapsing the container by one dimension, i.e. dense<T, 1> -> scalar
  result = add_reduce(c);
}


template<typename R, typename T>
void sum_of_squared_differences_v2(R& result, const dense<T>& a, const dense<T>& b)
{
  // copy-conversion c'tor to proceed with higher precision
  const dense<T> c = (a - b) * (a - b);

  // cast the source of reduction upwards in order to achieve higher precision
  // performance note: cast operator is "fused" into reduction process
  // C++ note: higher precision of lhs is *not* sufficient to cast upwards
  result = add_reduce(static_cast<dense<R> >(c));
}


int main(int /*argc*/, char* /*argv*/[])
{
  int ret_code;

  try {
    typedef f32 T; // type of the inputs
    typedef f64 R; // type of the result
    // corresponding built-in type
    typedef uncaptured<T>::type UT;

    const std::size_t size = 1024;
    std::vector<UT> data_a(size), data_b(size);

    const double scale = 1.0 / size;
    for (std::size_t i = 0; i != size; ++i) {
      data_a[i] = -static_cast<UT>(scale * i);
      data_b[i] = static_cast<UT>(scale * (size - i - 1));
    }

    // lifetime needs to be longer than the lifetime of the bound buffers
    dense<T> a, b; // empty dense containers
    bind(a, &data_a[0], size);
    bind(b, &data_b[0], size);

    R result; // scalar result (reduction)
    call(sum_of_squared_differences<R, T>)(result, a, b);

    std::cout << "Result: " << value(result) << '\n';

    ret_code = EXIT_SUCCESS;
  }
  catch(const std::exception& e) {
    std::cerr << "\n\nError: " << e.what() << '\n';
    ret_code = EXIT_FAILURE;
  }
  catch(...) {
    std::cerr << "\n\nError: Unknown exception caught!\n";
    ret_code = EXIT_FAILURE;
  }

  return ret_code;
}
