/****
***** Copyright 2010 Intel Corporation All Rights Reserved.
*****
***** The source code, information and material contained herein are owned by Intel Corporation or its suppliers  *****
***** or licensors, and title to such Material remains with Intel Corporation or its suppliers or licensors.      *****
***** The Material contains proprietary information of Intel or its suppliers and licensors. The Material is      *****
***** protected by worldwide copyright laws and treaty provisions. No part of the Material may be used, copied,   *****
***** reproduced, modified, published, uploaded, posted, transmitted, distributed or disclosed in any way without *****
***** Intel's prior express written permission.
*****
***** No license under any patent, copyright or other intellectual property rights in the material is granted to  *****
***** or conferred upon you, either expressly, by implication, inducement, estoppel or otherwise. Any license     *****
***** under such intellectual property rights must be express and approved by Intel in writing.
****/

/**** Copyright Ends ****/

#include <arbb.hpp>
#include <limits>


using namespace arbb;


template<typename R, typename T>
void squared_differences_cpp(R& result, const T& a, const T& b)
{
  for (std::size_t i = 0, end = a.size(); i != end; ++i) {
    result[i] = static_cast<typename R::value_type>((a[i] - b[i]) * (a[i] - b[i]));
  }
}


template<typename R, typename T>
void squared_differences_for(dense<R>& result, const dense<T>& a, const dense<T>& b)
{
  _for (usize i = 0, i != a.length(), ++i)
  {
    result[i] = static_cast<R>((a[i] - b[i]) * (a[i] - b[i]));
  }
  _end_for;
}


template<typename R, typename T>
void squared_differences_map(dense<R>& result, const dense<T>& a, const dense<T>& b)
{
  struct local {
    static void stencil(R& result, const T& a, const T& b) {
      result = static_cast<R>((a - b) * (a - b));
    }
  };

  map(local::stencil)(result, a, b);
}


int main(int /*argc*/, char* /*argv*/[])
{
  int ret_code;

  try {
    typedef f32 T;
    typedef f32 R;
    typedef uncaptured<T>::type UT;
    typedef uncaptured<R>::type UR;

    const std::size_t size = 1024 * 1024 * 8;
    std::vector<UT> data_a(size), data_b(size), data_c(size);

    const double scale = 1.0 / size;
    for (std::size_t i = 0; i != size; ++i) {
      data_a[i] = -static_cast<UT>(scale * i);
      data_b[i] = static_cast<UT>(scale * (size - i - 1));
    }

    dense<T> a, b, c;
    bind(a, &data_a[0], size);
    bind(b, &data_b[0], size);
    bind(c, &data_c[0], size);

    const std::size_t ncalls = 7;

    double time1 = std::numeric_limits<double>::max();
    for (std::size_t i = 0; i != ncalls; ++i) {
      double time_i;
      { // begin profiling section using scoped_timer
        const scoped_timer timer(time_i);
        squared_differences_cpp(data_c, data_a, data_b);
      } // end profiling section using scoped_timer
      time1 = std::min(time1, time_i);
    }
    std::cout << "Time/cpp: " << time1 << " ms\n";

    double time2 = std::numeric_limits<double>::max();
    for (std::size_t i = 0; i != ncalls; ++i) {
      double time_i;
      { // begin profiling section using scoped_timer
        const scoped_timer timer(time_i);
        call(squared_differences_for<R, T>)(c, a, b);
      } // end profiling section using scoped_timer
      time2 = std::min(time2, time_i);
    }
    std::cout << "Time/for: " << time2 << " ms\n";

    double time3 = std::numeric_limits<double>::max();
    for (std::size_t i = 0; i != ncalls; ++i) {
      double time_i;
      { // begin profiling section using scoped_timer
        const scoped_timer timer(time_i);
        call(squared_differences_map<R, T>)(c, a, b);
      } // end profiling section using scoped_timer
      time3 = std::min(time3, time_i);
    }
    std::cout << "Time/map: " << time3 << " ms\n";

    ret_code = EXIT_SUCCESS;
  }
  catch(const std::exception& e) {
    std::cerr << "\n\nError: " << e.what() << '\n';
    ret_code = EXIT_FAILURE;
  }
  catch(...) {
    std::cerr << "\n\nError: Unknown exception caught!\n";
    ret_code = EXIT_FAILURE;
  }

  return ret_code;
}
