/****
***** Copyright 2010 Intel Corporation All Rights Reserved.
*****
***** The source code, information and material contained herein are owned by Intel Corporation or its suppliers  *****
***** or licensors, and title to such Material remains with Intel Corporation or its suppliers or licensors.      *****
***** The Material contains proprietary information of Intel or its suppliers and licensors. The Material is      *****
***** protected by worldwide copyright laws and treaty provisions. No part of the Material may be used, copied,   *****
***** reproduced, modified, published, uploaded, posted, transmitted, distributed or disclosed in any way without *****
***** Intel's prior express written permission.
*****
***** No license under any patent, copyright or other intellectual property rights in the material is granted to  *****
***** or conferred upon you, either expressly, by implication, inducement, estoppel or otherwise. Any license     *****
***** under such intellectual property rights must be express and approved by Intel in writing.
****/

/**** Copyright Ends ****/

#include <arbb.hpp>
#include <limits>


using namespace arbb;


template<typename T>
void function(dense<T>& result, const dense<T>& a, const dense<T>& b)
{
  result = sin(a) + cos(b);
}


template<typename T>
void validate(arbb::boolean& result, const dense<T>& a, const dense<T>& b)
{
  result = all(a == b);
}


int main(int /*argc*/, char* /*argv*/[])
{
  int ret_code;

  try {
    typedef f64 T;
    typedef uncaptured<T>::type UT;
    typedef std::complex<T> CT;

    const std::size_t size = 1024 * 1024 * 4;
    std::vector<UT> areal(size), aimag(size);
    std::vector<UT> breal(size), bimag(size);
    std::vector<UT> creal(size), cimag(size);
    std::vector<std::complex<UT> > acplx(size), bcplx(size), ccplx(size);

    for (std::size_t i = 0; i != size; ++i) {
      const UT x = static_cast<UT>(i);
      areal[i] = x;
      aimag[i] = x / (i + 1);
      acplx[i] = std::complex<UT>(areal[i], aimag[i]);
      breal[i] = areal[i];
      bimag[i] = -aimag[i];
      bcplx[i] = std::complex<UT>(breal[i], bimag[i]);
    }

    dense<CT> a1, b1, c1;
    bind(a1, size, &areal[0], &aimag[0]);
    bind(b1, size, &breal[0], &bimag[0]);
    bind(c1, size, &creal[0], &cimag[0]);

    dense<CT> a2, b2, c2;
    // perform AoS to SoA conversion by specifying the pitch between member data
    const std::size_t pitch = sizeof(std::complex<UT>);
    void *adata = &acplx[0], *bdata = &bcplx[0], *cdata = &ccplx[0];
    bind(a2, size, adata, pitch, static_cast<unsigned char*>(adata) + sizeof(UT), pitch);
    bind(b2, size, bdata, pitch, static_cast<unsigned char*>(bdata) + sizeof(UT), pitch);
    bind(c2, size, cdata, pitch, static_cast<unsigned char*>(cdata) + sizeof(UT), pitch);

    const std::size_t ncalls = 7;
    double time1 = std::numeric_limits<double>::max();
    for (std::size_t i = 0; i != ncalls; ++i) {
      double time;
      { // begin profiling section using scoped_timer
        const scoped_timer timer(time);
        call(function<CT>)(c1, a1, b1);
        // make sure the bound buffer is up-to-date
        c1.read_only_range();
      } // end profiling section using scoped_timer
      time1 = std::min(time1, time);
    }
    std::cout << "Time 1: " << time1 << " ms\n";

    double time2 = std::numeric_limits<double>::max();
    for (std::size_t i = 0; i != ncalls; ++i) {
      double time;
      { // begin profiling section using scoped_timer
        const scoped_timer timer(time);
        call(function<CT>)(c2, a2, b2);
        // make sure the bound buffer is up-to-date
        c2.read_only_range();
      } // end profiling section using scoped_timer
      time2 = std::min(time2, time);
    }
    std::cout << "Time 2: " << time2 << " ms\n";

    // validation using an ArBB program
    arbb::boolean ok = false;
    call(validate<CT>)(ok, c1, c2);
    if (!value(ok)) throw std::runtime_error("Verification #1 failed!");

    // validation using C/C++
    for (std::size_t i = 0; i != size; ++i) {
      const std::complex<UT>& c = ccplx[i];
      if (c.real() != creal[i] || c.imag() != cimag[i]) {
        std::ostringstream oss;
        oss << "Verification #2 at position #" << (i + 1) << " failed!";
        throw std::runtime_error(oss.str());
      }
    }

    ret_code = EXIT_SUCCESS;
  }
  catch(const std::exception& e) {
    std::cerr << "\n\nError: " << e.what() << '\n';
    ret_code = EXIT_FAILURE;
  }
  catch(...) {
    std::cerr << "\n\nError: Unknown exception caught!\n";
    ret_code = EXIT_FAILURE;
  }

  return ret_code;
}
