/****
***** Copyright 2010 Intel Corporation All Rights Reserved.
*****
***** The source code, information and material contained herein are owned by Intel Corporation or its suppliers  *****
***** or licensors, and title to such Material remains with Intel Corporation or its suppliers or licensors.      *****
***** The Material contains proprietary information of Intel or its suppliers and licensors. The Material is      *****
***** protected by worldwide copyright laws and treaty provisions. No part of the Material may be used, copied,   *****
***** reproduced, modified, published, uploaded, posted, transmitted, distributed or disclosed in any way without *****
***** Intel's prior express written permission.
*****
***** No license under any patent, copyright or other intellectual property rights in the material is granted to  *****
***** or conferred upon you, either expressly, by implication, inducement, estoppel or otherwise. Any license     *****
***** under such intellectual property rights must be express and approved by Intel in writing.
****/

/**** Copyright Ends ****/

#include <arbb.hpp>
#include <limits>


using namespace arbb;


template<typename T>
class interval {
public:
  typedef T value_type;

  interval(): m_data(make_array<2, T>(0)) {}
  explicit interval(const T& t)     { m_data[0] = t; m_data[1] = t; }
  interval(const T& a, const T& b)  { m_data[0] = a; m_data[1] = b; }

  interval operator()(std::size_t i, std::size_t j) const {
    interval result;
    result.m_data[0] = m_data.at(i);
    result.m_data[1] = m_data.at(j);
    return result;
  }

  interval& operator+=(const interval& rhs) {
    m_data += rhs.m_data;
    return *this;
  }

  interval& operator-=(const interval& rhs) {
    m_data -= rhs(1, 0).m_data;
    return *this;
  }

  interval& operator*=(const interval& rhs) {
    array<data_type, 2> a = { m_data * rhs.m_data, m_data * rhs(1, 0).m_data };
    m_data[0] = min_reduce(min_reduce(a));
    m_data[1] = max_reduce(max_reduce(a));
    return *this;
  }

  // division by zero not handled properly (simple implementation)
  interval& operator/=(const interval& rhs) {
    array<data_type, 2> a = { m_data / rhs.m_data, m_data / rhs(1, 0).m_data };
    m_data[0] = min_reduce(min_reduce(a));
    m_data[1] = max_reduce(max_reduce(a));
    return *this;
  }

  T& lower()              { return m_data[0]; }
  const T& lower() const  { return m_data[0]; }

  T& upper()              { return m_data[1]; }
  const T& upper() const  { return m_data[1]; }

  // usually we do not need the absolute value (already positive)
  T width() const { return m_data[1] - m_data[0]; }

  // make width available as function in ArBB to work on whole containers
  ARBB_ELTWISE_TMETHOD_0(T, const interval<T>, width)

private:
  typedef array<T, 2> data_type;
  data_type m_data;
};

template<typename T> interval<T> operator+(const interval<T>& a, const interval<T>& b) { return interval<T>(a) += b; }
template<typename T> interval<T> operator-(const interval<T>& a, const interval<T>& b) { return interval<T>(a) -= b; }
template<typename T> interval<T> operator*(const interval<T>& a, const interval<T>& b) { return interval<T>(a) *= b; }
template<typename T> interval<T> operator/(const interval<T>& a, const interval<T>& b) { return interval<T>(a) /= b; }


template<typename T>
void function(dense<T>& result, const dense<T>& a, const dense<T>& b)
{
  dense<T> c = a * a;
  dense<T> d = (c + T(1)) / (c + T(2)); // non-zero
  result = (a * b) / d;
}


template<typename T>
void max_width(dense<T>& result, const dense<interval<T> >& a, const dense<interval<T> >& b)
{
  result = max(width(a), width(b));
}


template<typename T>
void get_width(dense<T>& result, const dense<interval<T> >& a)
{
  result = width(a);
}


template<typename T>
void total(T& result, const dense<T>& a)
{
  result = add_reduce(a);
}


int main(int /*argc*/, char* /*argv*/[])
{
  int ret_code;

  try {
    typedef f32 T;
    typedef interval<T> IT;
    typedef uncaptured<T>::type UT;

    const std::size_t size = 1024 * 1024 * 16;
    std::vector<interval<UT> > data_a(size), data_b(size);

    const UT nth = 1000 * static_cast<UT>(RAND_MAX);
    for (std::size_t i = 0; i != size; ++i) {
      const UT j = static_cast<UT>(i);
      const UT x = j / (i + 1);
      const UT y = j / (i + 2);
      data_a[i] = interval<UT>(x, x + std::rand() / nth);
      data_b[i] = interval<UT>(y, y + std::rand() / nth);
    }

    dense<IT> a, b;
    // perform AoS to SoA conversion by specifying the pitch between member data
    const std::size_t pitch = sizeof(interval<UT>);
    bind(a, size, &data_a[0].lower(), pitch, &data_a[0].upper(), pitch);
    bind(b, size, &data_b[0].lower(), pitch, &data_b[0].upper(), pitch);

    dense<T> w;
    call(max_width<T>)(w, a, b);

    T error0;
    call(total<T>)(error0, w);

    dense<IT> c;
    call(function<IT>)(c, a, b);

    T error1;
    call(get_width<T>)(w, c);
    call(total<T>)(error1, w);

    const UT e0 = value(error0), e1 = value(error1);
    std::cout << "Error (initial): " << e0 << '\n';
    std::cout << "Error (final)  : " << e1 << '\n';

    if (0 != e0) {
      std::cout << "Magnification  : " << (e1 / e0) << '\n';
    }

    ret_code = EXIT_SUCCESS;
  }
  catch(const std::exception& e) {
    std::cerr << "\n\nError: " << e.what() << '\n';
    ret_code = EXIT_FAILURE;
  }
  catch(...) {
    std::cerr << "\n\nError: Unknown exception caught!\n";
    ret_code = EXIT_FAILURE;
  }

  return ret_code;
}
