/****
***** Copyright 2010 Intel Corporation All Rights Reserved.
*****
***** The source code, information and material contained herein are owned by Intel Corporation or its suppliers  *****
***** or licensors, and title to such Material remains with Intel Corporation or its suppliers or licensors.      *****
***** The Material contains proprietary information of Intel or its suppliers and licensors. The Material is      *****
***** protected by worldwide copyright laws and treaty provisions. No part of the Material may be used, copied,   *****
***** reproduced, modified, published, uploaded, posted, transmitted, distributed or disclosed in any way without *****
***** Intel's prior express written permission.
*****
***** No license under any patent, copyright or other intellectual property rights in the material is granted to  *****
***** or conferred upon you, either expressly, by implication, inducement, estoppel or otherwise. Any license     *****
***** under such intellectual property rights must be express and approved by Intel in writing.
****/

/**** Copyright Ends ****/

#include <arbb.hpp>
#include <limits>


using namespace arbb;


/// Ordinary scalar product using a reduction.
void scalar_product(f32& result, const dense<f32>& a, const dense<f32>& b)
{
  result = add_reduce(a * b);
}


/// Matrix-vector product performing row-wise scalar products.
void matvec_product_for(dense<f32>& result, const dense<f32, 2>& matrix, const dense<f32>& vector)
{
  _for (usize i = 0, i != vector.length(), ++i) {
    f32 scalar_product_result;
    scalar_product(scalar_product_result, matrix.row(i), vector);
    result[i] = scalar_product_result;
  } _end_for;
}


/// Matrix-vector product using element-wise operations followed by a reduction along the columns.
void matvec_product(dense<f32>& result, const dense<f32, 2>& matrix, const dense<f32>& vector)
{
  _for (usize i = 0, i != vector.length(), ++i) {
    f32 scalar_product_result;
    scalar_product(scalar_product_result, matrix.row(i), vector);
    result[i] = scalar_product_result;
  } _end_for;
}


int main(int /*argc*/, char* /*argv*/[])
{
  int ret_code;

  try {
    const std::size_t size = 1024 * 4;
    std::vector<float> vdata(size), rdata(size);
    std::vector<float> mdata(size * size);

    for (std::size_t i = 0; i != size; ++i) {
      for (std::size_t j = 0; j != size; ++j) {
        mdata[i * size + j] = static_cast<float>(i + j) / (i + j + 1);
      }

      vdata[i] = static_cast<float>(i) / (size - i);
    }

    dense<f32> vector, result;
    dense<f32, 2> matrix;

    bind(vector, &vdata[0], size);
    bind(result, &rdata[0], size);
    bind(matrix, &mdata[0], size, size);

    const std::size_t ncalls = 7;

    double time_for = std::numeric_limits<double>::max();
    for (std::size_t i = 0; i != ncalls; ++i) {
      double time;

      { // begin profiling section using scoped_timer
        const scoped_timer timer(time);
        call(matvec_product_for)(result, matrix, vector);
        // make sure the bound buffer is up-to-date
        result.read_only_range();
      } // end profiling section using scoped_timer

      time_for = std::min(time_for, time);
      std::cout << "Time " << (i + 1) << "  : " << time << " ms\n";
    }
    std::cout << "Time/for: " << time_for << " ms\n";

    std::cout << "\n";

    double time_elt = std::numeric_limits<double>::max();
    for (std::size_t i = 0; i != ncalls; ++i) {
      double time;

      { // begin profiling section using scoped_timer
        const scoped_timer timer(time);
        call(matvec_product)(result, matrix, vector);
        // make sure the bound buffer is up-to-date
        result.read_only_range();
      } // end profiling section using scoped_timer

      time_elt = std::min(time_elt, time);
      std::cout << "Time " << (i + 1) << "  : " << time << " ms\n";
    }
    std::cout << "Time/for: " << time_elt << " ms\n";

    ret_code = EXIT_SUCCESS;
  }
  catch(const std::exception& e) {
    std::cerr << "\n\nError: " << e.what() << '\n';
    ret_code = EXIT_FAILURE;
  }
  catch(...) {
    std::cerr << "\n\nError: Unknown exception caught!\n";
    ret_code = EXIT_FAILURE;
  }

  return ret_code;
}
