/****
***** Copyright 2010 Intel Corporation All Rights Reserved.
*****
***** The source code, information and material contained herein are owned by Intel Corporation or its suppliers  *****
***** or licensors, and title to such Material remains with Intel Corporation or its suppliers or licensors.      *****
***** The Material contains proprietary information of Intel or its suppliers and licensors. The Material is      *****
***** protected by worldwide copyright laws and treaty provisions. No part of the Material may be used, copied,   *****
***** reproduced, modified, published, uploaded, posted, transmitted, distributed or disclosed in any way without *****
***** Intel's prior express written permission.
*****
***** No license under any patent, copyright or other intellectual property rights in the material is granted to  *****
***** or conferred upon you, either expressly, by implication, inducement, estoppel or otherwise. Any license     *****
***** under such intellectual property rights must be express and approved by Intel in writing.
****/

/**** Copyright Ends ****/

#include <arbb.hpp>
#include <limits>


using namespace arbb;


/// Matrix-vector product using element-wise operations followed by a reduction along the columns.
template<typename T>
void matvec_product(dense<T>& result, const dense<T, 2>& matrix, const dense<T>& vector)
{
  result = add_reduce(matrix * repeat_row(vector, matrix.num_rows()));
}


int main(int /*argc*/, char* /*argv*/[])
{
  int ret_code;

  try {
    typedef f32 T;
    typedef uncaptured<T>::type UT;

    const std::size_t size = 1024 * 4;
    std::vector<UT> vdata(size);
    std::vector<UT> mdata(size * size);

    for (std::size_t i = 0; i != size; ++i) {
      for (std::size_t j = 0; j != size; ++j) {
        mdata[i * size + j] = static_cast<UT>(i + j) / (i + j + 1);
      }

      vdata[i] = static_cast<UT>(i) / (size - i);
    }

    dense<f32> vector, result;
    dense<f32, 2> matrix;

    bind(vector, &vdata[0], size);
    bind(matrix, &mdata[0], size, size);

    // closure is strongly typed (signature/return type)
    closure<void(dense<T>&, const dense<T, 2>&, const dense<T>&)> c;

    double time0;
    { // profiling section using scoped_timer
      const scoped_timer timer(time0);
      c = capture(matvec_product<T>);
    }

    const std::size_t ncalls = 7;

    double time = std::numeric_limits<double>::max();
    for (std::size_t i = 0; i != ncalls; ++i) {
      double time_i;

      { // begin profiling section using scoped_timer
        const scoped_timer timer(time_i);
        c(result, matrix, vector);
        // make sure the bound buffer is up-to-date
        result.read_only_range();
      } // end profiling section using scoped_timer

      time = std::min(time, time_i);
      std::cout << "Time " << (i + 1) << ": " << time_i << " ms\n";
    }

    std::cout << "Capture Time: " << time0 << " ms\n";
    std::cout << "Compute Time: " << time << " ms\n";

    ret_code = EXIT_SUCCESS;
  }
  catch(const std::exception& e) {
    std::cerr << "\n\nError: " << e.what() << '\n';
    ret_code = EXIT_FAILURE;
  }
  catch(...) {
    std::cerr << "\n\nError: Unknown exception caught!\n";
    ret_code = EXIT_FAILURE;
  }

  return ret_code;
}
