/****
***** Copyright 2010 Intel Corporation All Rights Reserved.
*****
***** The source code, information and material contained herein are owned by Intel Corporation or its suppliers  *****
***** or licensors, and title to such Material remains with Intel Corporation or its suppliers or licensors.      *****
***** The Material contains proprietary information of Intel or its suppliers and licensors. The Material is      *****
***** protected by worldwide copyright laws and treaty provisions. No part of the Material may be used, copied,   *****
***** reproduced, modified, published, uploaded, posted, transmitted, distributed or disclosed in any way without *****
***** Intel's prior express written permission.
*****
***** No license under any patent, copyright or other intellectual property rights in the material is granted to  *****
***** or conferred upon you, either expressly, by implication, inducement, estoppel or otherwise. Any license     *****
***** under such intellectual property rights must be express and approved by Intel in writing.
****/

/**** Copyright Ends ****/

#include <arbb.hpp>
#include <limits>


using namespace arbb;


template<typename T>
void matvec_product(dense<T>& result, const dense<T, 2>& matrix, const dense<T>& vector)
{
  _for (usize i = 0, i != vector.length(), ++i) {
    result[i] = add_reduce(matrix.row(i) * vector);
  } _end_for;
}


template<typename T>
void matvec_product_unrolled(dense<T>& result, const dense<T, 2>& matrix, const dense<T>& vector)
{
  const std::size_t unroll_size = 8;

  const usize n = vector.length() / unroll_size;
  _for (usize i = 0, i != n, ++i) {
    for (std::size_t u = 0; u != unroll_size; ++u) {
      const usize j = i * unroll_size + u;
      result[j] = add_reduce(matrix.row(j) * vector);
    }
  } _end_for;
}


template<typename T>
void matvec_product_4096(dense<T>& result, const dense<T, 2>& matrix, const dense<T>& vector)
{
  _for (usize i = 0, i != vector.length(), ++i) {
    result[i] = add_reduce(matrix.row(i) * vector);
  } _end_for;
}


int main(int /*argc*/, char* /*argv*/[])
{
  int ret_code;

  try {
    typedef f32 T;
    typedef uncaptured<T>::type UT;

    const std::size_t size = 1024 * 4;
    std::vector<UT> vdata(size), rdata(size);
    std::vector<UT> mdata(size * size);

    for (std::size_t i = 0; i != size; ++i) {
      for (std::size_t j = 0; j != size; ++j) {
        mdata[i * size + j] = static_cast<UT>(i + j) / (i + j + 1);
      }

      vdata[i] = static_cast<UT>(i) / (size - i);
    }

    dense<f32> vector, result;
    dense<f32, 2> matrix;

    bind(vector, &vdata[0], size);
    bind(result, &rdata[0], size);
    bind(matrix, &mdata[0], size, size);

    // auto_closure is is runtime-checked
    const auto_closure ac1 = capture(matvec_product<T>);
    const auto_closure ac2 = capture(matvec_product_unrolled<T>);
    const auto_closure ac3 = capture(matvec_product_4096<T>);

    const std::size_t ncalls = 7;

    double time = std::numeric_limits<double>::max();
    for (std::size_t i = 0; i != ncalls; ++i) {
      double time_i;

      { // begin profiling section using scoped_timer
        const scoped_timer timer(time_i);
        ac1(result, matrix, vector);
        // make sure the bound buffer is up-to-date
        result.read_only_range();
      } // end profiling section using scoped_timer

      time = std::min(time, time_i);
      std::cout << "Time " << (i + 1) << ": " << time_i << " ms\n";
    }

    std::cout << "Time: " << time << " ms\n";

    // print the final result
    //std::copy(rdata.begin(), rdata.end(), std::ostream_iterator<UT>(std::cout, " "));

    ret_code = EXIT_SUCCESS;
  }
  catch(const std::exception& e) {
    std::cerr << "\n\nError: " << e.what() << '\n';
    ret_code = EXIT_FAILURE;
  }
  catch(...) {
    std::cerr << "\n\nError: Unknown exception caught!\n";
    ret_code = EXIT_FAILURE;
  }

  return ret_code;
}
