/****
***** Copyright 2010 Intel Corporation All Rights Reserved.
*****
***** The source code, information and material contained herein are owned by Intel Corporation or its suppliers  *****
***** or licensors, and title to such Material remains with Intel Corporation or its suppliers or licensors.      *****
***** The Material contains proprietary information of Intel or its suppliers and licensors. The Material is      *****
***** protected by worldwide copyright laws and treaty provisions. No part of the Material may be used, copied,   *****
***** reproduced, modified, published, uploaded, posted, transmitted, distributed or disclosed in any way without *****
***** Intel's prior express written permission.
*****
***** No license under any patent, copyright or other intellectual property rights in the material is granted to  *****
***** or conferred upon you, either expressly, by implication, inducement, estoppel or otherwise. Any license     *****
***** under such intellectual property rights must be express and approved by Intel in writing.
****/

/**** Copyright Ends ****/

#include "mimg.hpp"
#include "fit.hpp"
#include <arbb.hpp>


/// Function with a mixed signature of ArBB type and non-ArBB types.
/// C/C++ arguments are used to perform meta programming steps according
/// the state of these arguments, e.g. "radius" is "hardcoding" a specific
/// stencil pattern.
template<std::size_t D, std::size_t N, typename T, typename K>
void stencil(std::size_t radius, const arbb::array<arbb::usize, D>& size, arbb::array<T, N>& value)
{
  using namespace arbb;
  array<usize, D> i;
  // coordinate of the actual element of this "elemantal function" (stencil)
  position(i);

  _if (all(radius <= i && i < size - radius)) { // boundary condition
    array<T, N> a = make_array<N, T>(0);
    array<isize, D> offset = make_array<D, isize>(0);

    for (std::size_t d = 0; d != D; ++d) {
      for (int r = 1; r <= static_cast<int>(radius); ++r) {
        offset[d] = +r;
        a += neighbor(value, offset).template as<K>();
        offset[d] = -r;
        a += neighbor(value, offset).template as<K>();
      }

      offset[d] = 0; // reset
    }

    typedef typename uncaptured<K>::type UK;
    value = (a / static_cast<UK>(2 * D * radius)).template as<T>();
  } _end_if;
}


/// Store a stencil/kernel of a certain signature but still allow
/// to generate code statically considering the kernel radius.
template<std::size_t D, std::size_t N, typename T>
class heat_dissipation {
public:
  heat_dissipation(): m_radius(0) {}

  template<typename kernel_t>
  heat_dissipation(kernel_t kernel, std::size_t radius) {
    set_kernel<kernel_t>(kernel, radius);
  }

  /// Allows to (re-)set the kernel, and to trigger (re-)capturing IR.
  template<typename kernel_t>
  void set_kernel(kernel_t kernel, std::size_t radius) {
    m_radius = radius;

    arg<kernel_t, arg_kernel>() = kernel;
    arg<std::size_t, arg_radius>() = radius;

    m_captured = arbb::capture(&execute<kernel_t>);
  }

  void operator()(arbb::usize niterations, arbb::dense<arbb::array<T, N>, D>& grid) const {
    m_captured(niterations, grid);
  }

private:
  enum { arg_kernel, arg_radius };

  /// Helper to receive static class data according to the "id".
  template<typename arg_t, int id>
  static arg_t& arg() {
    static arg_t a;
    return a;
  }

  template<typename kernel_t>
  static void map_kernel(const arbb::array<arbb::usize, D>& size, arbb::array<T, N>& value) {
    // receive the static class data
    const kernel_t kernel = arg<kernel_t, arg_kernel>();
    const std::size_t radius = arg<std::size_t, arg_radius>();

    kernel(radius, size, value); // inline the code
  }

  template<typename kernel_t>
  static void execute(arbb::usize niterations, arbb::dense<arbb::array<T, N>, D>& grid) {
    using namespace arbb;
    array<usize, D> size = grid.size();

    _for (usize i = 0, i != niterations, ++i) {
      map(&map_kernel<kernel_t>)(size, grid);
    } _end_for;
  }

  arbb::auto_closure m_captured;
  std::size_t m_radius;
};


template<std::size_t D, std::size_t N, typename T> struct fire_ice {};

template<std::size_t N, typename T>
struct fire_ice<2, N, T> {
  fire_ice(arbb::array<T, N>* dst, const arbb::array<std::size_t, 2>& size, std::size_t radius)
  {
    using namespace arbb;
    const std::size_t linear_size = product(size);
    double scale = 1.0, shift = 0.0;
    fit<T>(0.0, 1.0, 0.0, 255.0, 0.0, 255.0, scale, shift);
    std::fill(dst, dst + linear_size, make_array<N, T>(static_cast<T>(0)));

    const std::size_t w = size[1];
    const std::size_t h = size[0];
    const double x = static_cast<double>(M_PI) / h;

    int maxr = static_cast<int>(radius);
    for (std::size_t i = 0; i != h; ++i) {
      for (int r = 0; r != maxr; ++r) {
        const T value = static_cast<T>(scale * std::sin(i * x) + shift);

        std::size_t j = r;
        dst[i * w + j].front() = value; // hot/red

        j = w - r - 1;
        dst[i * w + j].back() = value; // cold/blue
      }
    }
  }
};

template<std::size_t N, typename T>
struct fire_ice<3, N, T> {
  fire_ice(arbb::array<T, N>* dst, const arbb::array<std::size_t, 3>& size, std::size_t radius)
  {
    const std::size_t depth = size[2];
    const arbb::array<std::size_t, 2> size2 = arbb::resize<2>(size);
    const std::size_t page_size = product(size2);

    for (std::size_t i = 0; i != depth; ++i) {
      fire_ice<2, N, T>(dst, size2, radius);
      dst += page_size;
    }
  }
};


int main(int /*argc*/, char* /*argv*/[])
{
  int ret_code;

  try {
    using namespace arbb;
    const std::size_t D = 3;
    const std::size_t N = 3;
    typedef f32 K; // compute/kernel type
    typedef f32 T;

    const std::size_t radius = 4;

    const array<std::size_t, D> size = make_array<D, std::size_t>(64);
    const std::size_t linear_size = product(size);

    typedef uncaptured<T>::type UT;
    std::vector<array<UT, N> > data(linear_size);
    fire_ice<D, N, UT>(&data[0], size, radius);

    dense<array<T, N>, D> grid;
    bind(grid, &data[0][0], size);

    heat_dissipation<D, N, T> algorithm;
    algorithm.set_kernel(stencil<D, N, T, K>, radius);

    const usize niterations = 32;

    double time;
    { // begin profiling section using scoped_timer
      const scoped_timer timer(time);
      algorithm(niterations, grid);
    } // end profiling section using scoped_timer

    std::cout << "Time: " << time << " ms\n";

    const std::string filename("heat.mhd");
    if (!meta_image_write(filename, &size[0], 0, D, N, &data[0][0])) {
      std::ostringstream oss;
      oss << "Saving " << filename << " failed!";
      throw std::runtime_error(oss.str());
    }

    ret_code = EXIT_SUCCESS;
  }
  catch(const std::exception& e) {
    std::cerr << "\n\nError: " << e.what() << '\n';
    ret_code = EXIT_FAILURE;
  }
  catch(...) {
    std::cerr << "\n\nError: Unknown exception caught!\n";
    ret_code = EXIT_FAILURE;
  }

  return ret_code;
}
