/****
***** Copyright 2010 Intel Corporation All Rights Reserved.
*****
***** The source code, information and material contained herein are owned by Intel Corporation or its suppliers  *****
***** or licensors, and title to such Material remains with Intel Corporation or its suppliers or licensors.      *****
***** The Material contains proprietary information of Intel or its suppliers and licensors. The Material is      *****
***** protected by worldwide copyright laws and treaty provisions. No part of the Material may be used, copied,   *****
***** reproduced, modified, published, uploaded, posted, transmitted, distributed or disclosed in any way without *****
***** Intel's prior express written permission.
*****
***** No license under any patent, copyright or other intellectual property rights in the material is granted to  *****
***** or conferred upon you, either expressly, by implication, inducement, estoppel or otherwise. Any license     *****
***** under such intellectual property rights must be express and approved by Intel in writing.
****/

/**** Copyright Ends ****/

#ifndef MIMG_HPP
#define MIMG_HPP

#include <algorithm>
#include <typeinfo>
#include <fstream>
#include <sstream>
#include <cassert>
#include <vector>
#include <string>


////////////////////////////////////////////////////////////////////////////////
/// Meta Image Format
////////////////////////////////////////////////////////////////////////////////

/// Set of flags intended to signal various conditions
/// during reading or writing the file format.
enum meta_image_flags {
  /// Default set of flags.
  meta_image_flag_default = 0x00000000,
  /// Format warnings encountered.
  meta_image_flag_warning = 0x00000001,
  /// Byte order is MSB first.
  meta_image_flag_MSB     = 0x00000002,
};


/// Type name fall-back function.
template<typename T> const char* meta_image_typename() { return 0; }

template<> inline const char* meta_image_typename<char>()           { return "MET_CHAR"; }
template<> inline const char* meta_image_typename<unsigned char>()  { return "MET_UCHAR"; }
template<> inline const char* meta_image_typename<short>()          { return "MET_SHORT"; }
template<> inline const char* meta_image_typename<unsigned short>() { return "MET_USHORT"; }
template<> inline const char* meta_image_typename<int>()            { return "MET_INT"; }
template<> inline const char* meta_image_typename<unsigned int>()   { return "MET_UINT"; }
template<> inline const char* meta_image_typename<long>()           { return "MET_LONG"; }
template<> inline const char* meta_image_typename<unsigned long>()  { return "MET_ULONG"; }
template<> inline const char* meta_image_typename<float>()          { return "MET_FLOAT"; }
template<> inline const char* meta_image_typename<double>()         { return "MET_DOUBLE"; }


/// Parse the header of a file.
/// The header can be part of the data file (local),
/// or separately stored (mha/mhd).
/// Returns the type of the data (typeinfo).
inline const std::type_info*
meta_image_read_header(
  const std::string& header_filename,
  std::string& filename,
  std::vector<std::size_t>& size,
  /// Number of components.
  std::size_t& ncomponents,
  int& flags,
  std::size_t* extension_size = 0,
  /// Local Meta Image format detected (if non-zero);
  /// file is containing header information and data.
  std::size_t* local_skip = 0,
  std::vector<double>* spacing = 0)
{
  std::ifstream file(header_filename.c_str(), std::ios::binary);
  if (!file) return 0;

  std::size_t ndims = 0;
  std::string data_type;

  size.clear();
  filename.clear();
  ncomponents = 1;
  flags = meta_image_flag_default;

  if (extension_size) {
    *extension_size = 0;
  }

  if (local_skip) {
    *local_skip = 0;
  }

  if (spacing) {
    spacing->clear();
  }

  std::string buffer;
  while (std::getline(file, buffer)) {
    std::stringstream ss;
    std::string buffer2;
    ss << buffer;
    ss >> buffer >> buffer2;

    if ("NDims" == buffer
      && 1 == buffer2.size() && '=' == *buffer2.begin())
    {
      ss >> ndims;
    }
    else if ("ElementNumberOfChannels" == buffer
      && 1 == buffer2.size() && '=' == *buffer2.begin())
    {
      ss >> ncomponents;
    }
    else if ("DimSize" == buffer
      && 1 == buffer2.size() && '=' == *buffer2.begin())
    {
      std::size_t u;
      while (ss >> u) {
        size.push_back(u);
      }
    }
    else if ("ElementType" == buffer
      && 1 == buffer2.size() && '=' == *buffer2.begin())
    {
      ss >> data_type;
    }
    else if (spacing && "ElementSpacing" == buffer
      && 1 == buffer2.size() && '=' == *buffer2.begin())
    {
      double d;
      while (ss >> d) {
        if (0.0 < d) {
          spacing->push_back(d);
        } else {
          // invalid spacing encountered
          return 0;
        }
      }
    }
    else if (("BinaryDataByteOrderMSB" == buffer || "ElementByteOrderMSB" == buffer)
      && 1 == buffer2.size() && '=' == *buffer2.begin())
    {
      ss >> buffer2;
      if ("False" == buffer2) {
        flags &= ~meta_image_flag_MSB;
      }
      else if ("True" == buffer2) {
        flags |= meta_image_flag_MSB;
      }
      else {
        flags |= meta_image_flag_warning;
      }
    }
    else if (extension_size && "ExtensionDataSize" == buffer
      && 1 == buffer2.size() && '=' == *buffer2.begin())
    {
      std::size_t dummy;
      if (ss >> dummy && 0 < dummy) {
        *extension_size = dummy;
      }
      else {
        flags |= meta_image_flag_warning;
      }
    }
    else if ("ElementDataFile" == buffer
      && 1 == buffer2.size() && '=' == *buffer2.begin())
    {
      ss >> buffer2;
      if ("LOCAL" == buffer2 || header_filename == buffer2) {
        if (local_skip) {
          *local_skip = static_cast<std::size_t>(file.tellg());
          filename = header_filename;
          // supposed to be the last statement before the raw data
          break;
        }
      }
      else {
        filename = buffer2;
      }
    }
    else {
      flags |= meta_image_flag_warning;
    }
  }

  // prefix the path of the header file to actually find the data file
  if (!filename.empty() // prevent creating only a path (without filename)
    // quick hint to pass this block
    && (!local_skip || 0 == *local_skip))
  {
    const std::size_t pos = header_filename.find_last_of("/\\");
    if (std::string::npos != pos) {
      // include the path separator: pos + 1
      std::string(header_filename.begin(), header_filename.begin() + pos + 1).append(filename).swap(filename);
    }
  }

  const std::type_info* type_info = 0;

  // perform a basic pre-validation before type-detection is performed
  if (size.size() == ndims && !filename.empty() && (!spacing || spacing->size() == ndims)) {
    if (meta_image_typename<char>() == data_type) {
      type_info = &typeid(char);
    }
    else if (meta_image_typename<unsigned char>() == data_type) {
      type_info = &typeid(unsigned char);
    }
    else if (meta_image_typename<short>() == data_type) {
      type_info = &typeid(short);
    }
    else if (meta_image_typename<unsigned short>() == data_type) {
      type_info = &typeid(unsigned short);
    }
    else if (meta_image_typename<int>() == data_type) {
      type_info = &typeid(int);
    }
    else if (meta_image_typename<unsigned int>() == data_type) {
      type_info = &typeid(unsigned int);
    }
    else if (meta_image_typename<long>() == data_type) {
      type_info = &typeid(long);
    }
    else if (meta_image_typename<unsigned long>() == data_type) {
      type_info = &typeid(unsigned long);
    }
    else if (meta_image_typename<float>() == data_type) {
      type_info = &typeid(float);
    }
    else if (meta_image_typename<double>() == data_type) {
      type_info = &typeid(double);
    }
  }

  return type_info;
}


/// Load the data file optionally allowing conversion.
/// Conversion is performed unconditionally, i.e. data
/// is clamped eventually to fit into the destination.
template<typename T>
bool
meta_image_read(
  /// Filename referring to the data.
  const std::string& filename,
  /// Real extents implicitly determining the dimensionality.
  const std::vector<std::size_t>& data_size,
  /// Requested extents of the data.
  const std::size_t* size,
  /// Requested dimensionality (number of entries in size).
  std::size_t ndims,
  /// Number of components.
  std::size_t ncomponents,
  /// Used to skip header information to only read the data.
  std::size_t local_skip,
  /// Or'ed flags.
  int flags,
  /// Buffer where the data is read into.
  T* data,
  /// Storage type required if not typeid(T);
  /// performs conversion from typo_info to T.
  const std::type_info* type_info = 0,
  /// Callback executed per entry during reading the data.
  /// Allows assigning the value to the left-hand-side;
  /// allows to validate the value against a given buffer.
  bool (*handle_entry)(T&, T) = 0,
  char* extension_begin = 0,
  char* extension_end = 0)
{
  assert(!filename.empty() && 0 != static_cast<const char*>("Empty filename cannot refer to file!"));
  assert(!data_size.empty() && 0 != static_cast<const char*>("Dimensionality cannot be zero!"));
  assert(0 != ndims && 0 != static_cast<const char*>("Dimensionality cannot be zero!"));
  assert(0 != ncomponents && 0 != static_cast<const char*>("Number of channels cannot be zero!"));
  assert(data && 0 != static_cast<const char*>("NULL-pointer cannot refer data!"));
  assert(extension_begin <= extension_end && 0 != static_cast<const char*>("Invalid extension range specified!"));

  std::ifstream file(filename.c_str(), std::ios::binary);
  bool ok = 0 == (flags & meta_image_flag_MSB) && 0 != file;

  if (ok && 0 != local_skip) {
    file.seekg(local_skip);
    ok = 0 != file;
  }

  if (ok) {
    // real data size
    std::size_t linear_data_size = data_size.empty() ? 0 : ncomponents * data_size[0];
    for (std::size_t i = 1, end = data_size.size(); i < end; ++i) {
      linear_data_size *= data_size[i];
    }

    // requested size
    std::size_t linear_size;
    if (size) {
      linear_size = ncomponents * size[0];
      for (std::size_t i = 1; i < ndims; ++i) {
        linear_size *= size[i];
      }
    }
    else {
      linear_size = linear_data_size;
    }

    const std::size_t streamsize = std::min(linear_data_size, linear_size);

    struct helper {
      static char* cast(void* dst) { return reinterpret_cast<char*>(dst); }
      static char* cast(const void* /*dummy*/) { return 0; }

      static bool assign(T& lhs, T rhs) {
        std::copy(cast(&rhs), cast(&rhs) + sizeof(T), cast(&lhs));
        return true;
      }
    };

    char *const destination = helper::cast(data);

    if (!handle_entry) {
      handle_entry = &helper::assign;
    }

    if ((type_info && typeid(T) != *type_info)
      // special handling required
      || &helper::assign != handle_entry)
    {
      if (*type_info == typeid(char)) {
        for (std::size_t i = 0; i != streamsize && ok; ++i) {
          char value;
          file.read(&value, static_cast<std::streamsize>(1));
          ok = handle_entry(data[i], static_cast<T>(value));
        }
      }
      else if (*type_info == typeid(unsigned char)) {
        for (std::size_t i = 0; i != streamsize && ok; ++i) {
          unsigned char value;
          file.read(reinterpret_cast<char*>(&value), static_cast<std::streamsize>(1));
          ok = handle_entry(data[i], static_cast<T>(value));
        }
      }
      else if (*type_info == typeid(short)) {
        for (std::size_t i = 0; i != streamsize && ok; ++i) {
          short value;
          file.read(reinterpret_cast<char*>(&value), static_cast<std::streamsize>(sizeof(short)));
          ok = handle_entry(data[i], static_cast<T>(value));
        }
      }
      else if (*type_info == typeid(unsigned short)) {
        for (std::size_t i = 0; i != streamsize && ok; ++i) {
          unsigned short value;
          file.read(reinterpret_cast<char*>(&value), static_cast<std::streamsize>(sizeof(unsigned short)));
          ok = handle_entry(data[i], static_cast<T>(value));
        }
      }
      else if (*type_info == typeid(float)) {
        for (std::size_t i = 0; i != streamsize && ok; ++i) {
          float value;
          file.read(reinterpret_cast<char*>(&value), static_cast<std::streamsize>(sizeof(float)));
          ok = handle_entry(data[i], static_cast<T>(value));
        }
      }
      else if (*type_info == typeid(double)) {
        for (std::size_t i = 0; i != streamsize && ok; ++i) {
          double value;
          file.read(reinterpret_cast<char*>(&value), static_cast<std::streamsize>(sizeof(double)));
          ok = handle_entry(data[i], static_cast<T>(value));
        }
      }
      else { // unsupported type
        ok = false;
      }
    }
    else { // fast path
      assert(destination && 0 != static_cast<const char*>("Cannot write to a constant destination!"));
      file.read(destination, static_cast<std::streamsize>(streamsize * sizeof(T)));
    }

    if (ok) {
      // fill the part which exceeds the real amount of data
      if (linear_data_size < linear_size && destination) {
        std::fill( // zero-out the destination
          destination + linear_data_size * sizeof(T),
          destination + linear_size * sizeof(T), 0);
      }

      const std::size_t extension_size = std::distance(extension_begin, extension_end);
      if (0 < extension_size) {
        file.read(extension_begin, static_cast<std::streamsize>(extension_size));
      }

      // basic stream validation
      file.clear(std::ios::eofbit);
      ok = 0 != file;
    }
  }

  return ok;
}


/// Save the file.
template<typename T>
bool
meta_image_write(
  const std::string& filename,
  const std::size_t* size,
  const double* spacing,
  std::size_t ndims,
  std::size_t ncomponents,
  const T* data,
  const char* extension_begin = 0,
  const char* extension_end = 0)
{
  assert(!filename.empty() && 0 != static_cast<const char*>("Empty filename cannot refer to file!"));
  assert(extension_begin <= extension_end && 0 != static_cast<const char*>("Invalid extension range specified!"));

  std::ofstream file(filename.c_str(), std::ios_base::binary);

  const char *const type = meta_image_typename<T>(); // translate type to type name
  bool ok = 0 != size && 0 != ndims && 0 != ncomponents && 0 != data && 0 != type && 0 != file;

  if (ok) { // perform basic validation
    file << "NDims = " << ndims << '\n';
    file << "ElementNumberOfChannels = " << ncomponents << '\n';
    file << "ElementByteOrderMSB = False\n";
    file << "ElementType = " << type << '\n';

    file << "DimSize =";
    for (std::size_t i = 0; i != ndims; ++i) {
      file << ' ' << size[i];
    }
    file << '\n';

    file << "ElementSpacing =";
    for (std::size_t i = 0; i != ndims; ++i) {
      file << ' ' << (spacing ? spacing[i] : 1.0);
    }
    file << '\n';

    // number of items silently append after the regular data section
    const std::size_t extension_size = std::distance(extension_begin, extension_end);
    if (0 < extension_size) {
      file << "ExtensionDataSize = " << extension_size << '\n';
    }

    // must be the last entry before writing the data
    file << "ElementDataFile = LOCAL\n";

    // calculate the linear size
    std::size_t linear_size = size ? *size : 0;
    for (std::size_t i = 1; i < ndims; ++i) {
      linear_size *= size[i];
    }

    file.write(reinterpret_cast<const char*>(data), ncomponents * sizeof(T) * linear_size);

    // append the extension data behind the regular data section
    if (0 < extension_size) {
      file.write(extension_begin, extension_size);
    }
  }

  return ok && 0 != file;
}

#endif // MIMG_HPP
