#include <iostream>
#include <arbb.hpp>


int main(int /*argc*/, char* /*argv*/[])
{
  int ret_code;

  try {
    using namespace arbb;

    arbb_context_t context;
    arbb_get_default_context(&context, detail::throw_on_error_details());

    arbb_type_t type;
    arbb_get_scalar_type(context, &type, arbb_f32, detail::throw_on_error_details());

    arbb_type_t inputs[]  = { type, type };
    arbb_type_t outputs[] = { type };

    arbb_type_t fn_type;
    arbb_get_function_type(context, &fn_type,
      sizeof(outputs) / sizeof(*outputs), outputs,
      sizeof(inputs)  / sizeof(*inputs), inputs,
      detail::throw_on_error_details());

    arbb_function_t function;
    {
      arbb_begin_function(context, &function, fn_type, "add", 0, detail::throw_on_error_details());
        arbb_variable_t a, b, c;
        enum { is_input, is_output };
        arbb_get_parameter(function, &a, is_input,  0, detail::throw_on_error_details());
        arbb_get_parameter(function, &b, is_input,  1, detail::throw_on_error_details());
        arbb_get_parameter(function, &c, is_output, 0, detail::throw_on_error_details());

        arbb_variable_t in[] = { a, b };
        arbb_variable_t out[] = { c };
        arbb_op(function, arbb_op_add, out, in, 0, detail::throw_on_error_details());
      arbb_end_function(function, detail::throw_on_error_details());
    }

    float data[] = { 25.0f, 7.0f };
    arbb_binding_t null_binding;
    arbb_set_binding_null(&null_binding);

    arbb_global_variable_t ga, gb, gc;
    arbb_create_constant(context, &ga, type, data + 0, 0, detail::throw_on_error_details());
    arbb_create_constant(context, &gb, type, data + 1, 0, detail::throw_on_error_details());

    arbb_variable_t in[2];
    arbb_get_variable_from_global(context, in + 0, ga, detail::throw_on_error_details());
    arbb_get_variable_from_global(context, in + 1, gb, detail::throw_on_error_details());

    arbb_variable_t out[1];
    arbb_create_global(context, &gc, type, "result", null_binding, 0, detail::throw_on_error_details());
    arbb_get_variable_from_global(context, out + 0, gc, detail::throw_on_error_details());

    double compile_time = 0;
    {
      const scoped_timer timer(compile_time, scoped_timer::unit_us);
      arbb_compile(function, detail::throw_on_error_details());
    }

    double compute_time = 0;
    const std::size_t ncalls = 100000;
    {
      const scoped_timer timer(compute_time);
      for (std::size_t i = 0; i != ncalls; ++i) {
        arbb_execute(function, out, in, detail::throw_on_error_details());
      }
    }

    float result = 0;
    arbb_read_scalar(context, out[0], &result, detail::throw_on_error_details());

    std::cout << "Compile time: " << compile_time << " us\n";
    std::cout << "Compute time: " << compute_time << " ms\n";
    std::cout << "Result      : " << result << '\n';

    ret_code = EXIT_SUCCESS;
  }
  catch(const std::exception& e) {
    std::cerr << "\n\nError: " << e.what() << '\n';
    ret_code = EXIT_FAILURE;
  }
  catch(...) {
    std::cerr << "\n\nError: Unknown exception caught!\n";
    ret_code = EXIT_FAILURE;
  }

  return ret_code;
}